-- Predictive Image Codec
-- /Mic 2002
--
-- Uses the FELICS algorithm described by Paul G. Howard and Jeffrey Scott Vitter.
-- Only intraband prediction is used for RGB images.
--

without warning


constant PIC_POWERSOF2 = {1,2,4,8,16,32,64,128,256,512,1024,2048,4096}
constant PIC_MAX_K = 7

constant IN_RANGE = 0,
	 OUT_OF_RANGE = 1,
	 BELOW_RANGE = 0,
	 ABOVE_RANGE = 1

sequence PIC_CEILLOG2

integer sbp_fn
integer buffer_,bitsinbuffer_,totalsize_
buffer_ = 0
bitsinbuffer_ = 0
totalsize_ = 0


procedure output_code(sequence C)
	for i=1 to length(C) do
		buffer_ = or_bits(buffer_,C[i]*PIC_POWERSOF2[bitsinbuffer_+1])
		bitsinbuffer_ += 1
		if bitsinbuffer_=8 then
			puts(sbp_fn,buffer_)
			totalsize_ += 1
			buffer_ = 0
			bitsinbuffer_ = 0
		end if
	end for
end procedure

function get_bit()
	integer b
	if bitsinbuffer_=0 then
		buffer_ = getc(sbp_fn)
		if buffer_=-1 then
			return -1
		end if
		bitsinbuffer_ = 8
	end if
	bitsinbuffer_ -= 1
	return (and_bits(buffer_,PIC_POWERSOF2[8-bitsinbuffer_])!=0)
end function


procedure flush()
	if bitsinbuffer_ then
		puts(sbp_fn,buffer_)
		totalsize_ += 1
	end if
end procedure



-- Return the Golomb-Rice code of non-negative integer n with parameter k.
-- The returned value is a sequence of 1's and 0's.
function grice(integer n,integer k)
	integer unary_part,binary_part
	sequence code
	
	unary_part = floor(n/PIC_POWERSOF2[k+1])
	code = repeat(1,unary_part) & 0
	binary_part = and_bits(n,PIC_POWERSOF2[k+1]-1)
	
	for i=k to 1 by -1 do
		code &= (and_bits(binary_part,PIC_POWERSOF2[i])!=0)
	end for
	
	return code
end function


function grice_decode(integer k)
	integer b,prefix,bin
	
	b = 1
	prefix = 0
	bin = 0
	
	while b != 0 do
		b = get_bit()
		if b=-1 then
			return -1
		end if
		prefix += b
	end while
	for i=k to 1 by -1 do
		bin += get_bit()*PIC_POWERSOF2[i]
	end for
	return bin+(prefix*PIC_POWERSOF2[k+1])
end function


procedure make_ceillog2_table()
	PIC_CEILLOG2 = repeat(1,257)
	for n=0 to 255 do
		for i=7 to 0 by -1 do
			if and_bits(n,PIC_POWERSOF2[i+1]) then
				PIC_CEILLOG2[n+1] = i+1
				exit
			end if
		end for
	end for
	PIC_CEILLOG2[257] = 9
end procedure
make_ceillog2_table()


function max(integer x,integer y)
	if x>y then
		return x
	end if
	return y
end function





function pic_encode_rgb(sequence image)
	atom pointer
	integer width,height
	integer offs,k,min_k
	integer H,L,delta,range,diff,
		log2range,threshold,longcodes,codelength
	sequence X,N,W
	sequence cumul_params,params

	pointer = image[1]
	width = image[2]
	height = image[3]

	cumul_params = repeat(repeat(0,8),256)
	-- Give initial penalties for using low k:s
	for i=0 to 255 do
		for j=0 to 7 do
			cumul_params[i+1][j+1] = floor((7-j)*(i/4))
		end for
	end for
	
	offs = 0
	
	for y=0 to height-1 do
		for x=0 to width-1 do
			X = peek({pointer+offs,3})
			if (y=0) then
				N = {192,192,192}
			else
				N = peek({pointer+offs-(width*3),3})
			end if
			if (x=0) then
				W = {64,64,64}
			else
				W = peek({pointer+offs-3,3})
			end if
			
			for i=1 to 3 do
				H = max(W[i],N[i])
				L = W[i]+N[i]-H
				
				delta = H-L
				range = delta + 1

				if (X[i]>=L) and (X[i]<=H) then
	    				output_code({IN_RANGE})
	    				X[i] -= L
	    				if range>1 then
	    					log2range = PIC_CEILLOG2[range+1]
	    					threshold = PIC_POWERSOF2[log2range+1] - range
	    					longcodes = range - threshold

	    					X[i] -= floor(longcodes/2)
	    					if X[i]<0 then
	    						X[i] += range
	    					end if
	    					if X[i]<threshold then
	    						codelength = log2range - 1
	    					else
	    						codelength = log2range
	    						X[i] += threshold
	    					end if
	    					for j=codelength to 1 by -1 do
	    						output_code({and_bits(X[i],PIC_POWERSOF2[j])!=0})
	    					end for
	      				end if
	      			else
	      				output_code({OUT_OF_RANGE})
					if X[i]<L then
						output_code({BELOW_RANGE})
						diff = L - X[i] - 1
					else
						output_code({ABOVE_RANGE})
						diff = X[i] - H - 1
					end if
					k = 0
					params = cumul_params[delta+1]
					min_k = params[1]
					for j=2 to PIC_MAX_K do
						if params[j] < min_k then
							min_k = params[j]
							k = j-1
						end if
					end for
					
					output_code(grice(diff,k))

					for j=1 to PIC_MAX_K do
						cumul_params[delta+1][j] += floor(diff/PIC_POWERSOF2[j])+j
					end for
				end if
			end for
			offs += 3
		end for
	end for
	flush()
	
	return totalsize_
end function


-- Encode an indexed image (ie. an 8-bit paletted image)
function pic_encode_idx(sequence image)
	atom pointer
	integer width,height
	integer offs,k,min_k
	integer H,L,delta,range,diff,
		log2range,threshold,longcodes,codelength
	integer X,N,W
	sequence cumul_params,params

	pointer = image[1]
	width = image[2]
	height = image[3]

	cumul_params = repeat(repeat(0,8),256)
	-- Give initial penalties for using low k:s
	for i=0 to 255 do
		for j=0 to 7 do
			cumul_params[i+1][j+1] = floor((7-j)*(i/4))
		end for
	end for	
	offs = 0
	
	for y=0 to height-1 do
		for x=0 to width-1 do
			
			X = peek(pointer+offs)
			if (y=0) then
				N = 192
			else
				N = peek(pointer+offs-width)
			end if
			if (x=0) then
				W = 64
			else
				W = peek(pointer+offs-1)
			end if

			H = max(W, N)
			L = W + N - H
				
			delta = H - L
			range = delta + 1
				
			if (X>=L) and (X<=H) then
    				output_code({IN_RANGE})
    				X -= L
    				if range>1 then
    					log2range = PIC_CEILLOG2[range+1]
    					threshold = PIC_POWERSOF2[log2range+1] - range
    					longcodes = range - threshold
    					X -= floor(longcodes/2)
    					if X<0 then
    						X += range
    					end if
    					if X<threshold then
    						codelength = log2range - 1
    					else
    						codelength = log2range
    						X += threshold
    					end if
    					for j=codelength to 1 by -1 do
    						output_code({and_bits(floor(X/PIC_POWERSOF2[j]),1)})
    					end for
      				end if
      			else
      				output_code({OUT_OF_RANGE})
				if X<L then
					output_code({BELOW_RANGE})
					diff = L - X - 1
				else
					output_code({ABOVE_RANGE})
					diff = X - H - 1
				end if
				k = 0
				params = cumul_params[delta+1]
				min_k = params[1]
				for j=2 to PIC_MAX_K do
					if params[j] < min_k then
						min_k = params[j]
						k = j-1
					end if
				end for
				
				output_code(grice(diff,k))
				for j=1 to PIC_MAX_K do
					cumul_params[delta+1][j] += floor(diff/PIC_POWERSOF2[j])+j
				end for
			end if
			offs += 1
		end for		
	end for
	flush()
	
	return totalsize_
end function


-- Encodes an image and writes the result to file 'fn'.
-- The 'image' sequence is assumed to be on the form {pointer,width,height,bitcount}
global function pic_encode(sequence image,integer fn)
	
	sbp_fn = fn
	bitsinbuffer_ = 0
	totalsize_ = 0
	
	if image[4] = 8 then
		return pic_encode_idx(image)
	elsif image[4] = 24 then
		return pic_encode_rgb(image)
	else
		return 0
	end if
end function



function pic_decode_rgb(sequence image)
	integer targetsize
	integer width,height,offs,b
	integer L,H,delta,range,log2range,
		longcodes,threshold,
		diff,k,min_k,which_side
	atom pointer
	sequence cumul_params
	sequence X,W,N
	
	pointer = image[1]
	width = image[2]
	height = image[3]
	targetsize = width*height*3
	offs = 0

	
	cumul_params = repeat(repeat(0,8),256)
	for i=0 to 255 do
		for j=0 to 7 do
			cumul_params[i+1][j+1] = floor((7-j)*(i/4))
		end for
	end for
	
	for y=0 to height-1 do
		for x=0 to width-1 do
			if y=0 then
				N = {192,192,192}
			else
				N = peek({pointer+offs-(width*3),3})
			end if
			
			if x=0 then
				W = {64,64,64}
			else
				W = peek({pointer+offs-3,3})
			end if

			
			X = {0,0,0}
			for i=1 to 3 do
				H = max(N[i],W[i])
				L = N[i]+W[i]-H
				delta = H-L
				range = delta+1
				
				if (get_bit() = IN_RANGE) then
			
					if range>1 then
	    					log2range = PIC_CEILLOG2[range+1]
	    					threshold = PIC_POWERSOF2[log2range+1] - range
 					
	    					for j=1 to log2range-1 do
	    						X[i] = (X[i]*2) + get_bit()
	    					end for
	    					if X[i]>=threshold then
	    						X[i] = (X[i]*2)+get_bit()
	    						X[i] -= threshold
	    					else
	    					end if
	    					longcodes = range - threshold
	    					X[i] += floor(longcodes/2)
	    					if X[i]>=range then
	    						X[i] -= range
	    					end if
	    					X[i] += L
	    					
					else
						X[i] = L
					end if

				else
					which_side = get_bit()
					k = 0
					min_k = cumul_params[delta+1][1]
					for j=2 to PIC_MAX_K do
						if cumul_params[delta+1][j]<min_k then
							min_k = cumul_params[delta+1][j]
							k = j-1
						end if
					end for
					
					diff = grice_decode(k)
					if diff=-1 then
						return offs
					end if
					
					for j=1 to PIC_MAX_K do
						cumul_params[delta+1][j] += floor(diff/PIC_POWERSOF2[j])+j
					end for
					
					if which_side=BELOW_RANGE then
						X[i] = L - diff - 1
					else
						X[i] = diff + H + 1
					end if
				end if
			end for
			poke(pointer+offs,X)
			offs += 3
		end for
	end for

	return offs
end function


function pic_decode_idx(sequence image)
	integer targetsize
	integer width,height,offs,b
	integer L,H,delta,range,log2range,
		longcodes,threshold,
		diff,k,min_k,which_side
	atom pointer
	sequence cumul_params
	integer X,W,N
	
	pointer = image[1]
	width = image[2]
	height = image[3]
	offs = 0

	
	cumul_params = repeat(repeat(0,8),256)
	for i=0 to 255 do
		for j=0 to 7 do
			cumul_params[i+1][j+1] = floor((7-j)*(i/4))
		end for
	end for
	
	for y=0 to height-1 do
		for x=0 to width-1 do
			if y=0 then
				N = 192
			else
				N = peek(pointer+offs-width)
			end if
			if x=0 then
				W = 64
			else
				W = peek(pointer+offs-1)
			end if
			
			X = 0
			H = max(N,W)
			L = N + W - H
			delta = H - L
			range = delta + 1 
				
			if (get_bit() = IN_RANGE) then
				if range>1 then
    					log2range = PIC_CEILLOG2[range+1]
    					threshold = PIC_POWERSOF2[log2range+1] - range
					
    					for j=1 to log2range-1 do
    						X = (X*2) + get_bit()
    					end for
    					if X>=threshold then
    						X = (X*2)+get_bit()
    						X -= threshold
    					else
    					end if
    					longcodes = range - threshold
    					X += floor(longcodes/2)
    					if X>=range then
    						X -= range
    					end if
    					X += L
				else
					X = L
				end if

			else
				which_side = get_bit()
				k = 0
				min_k = cumul_params[delta+1][1]
				for j=2 to PIC_MAX_K do
					if cumul_params[delta+1][j]<min_k then
						min_k = cumul_params[delta+1][j]
						k = j-1
					end if
				end for
				
				diff = grice_decode(k)
				if diff=-1 then
					return offs
				end if
				
				for j=1 to PIC_MAX_K do
					cumul_params[delta+1][j] += floor(diff/PIC_POWERSOF2[j])+j
				end for
				
				if which_side = BELOW_RANGE then
					X = L - diff - 1
				else
					X = diff + H + 1
				end if
			end if
			poke(pointer+offs,X)
			offs += 1
		end for
	end for

	return offs
end function


-- Decodes the encoded data in file 'fn' according to the values in 'image'.
global function pic_decode(integer fn,sequence image)
	
	sbp_fn = fn
	bitsinbuffer_ = 0
	totalsize_ = 0
	
	if image[4]=8 then
		return pic_decode_idx(image)
	elsif image[4]=24 then
		return pic_decode_rgb(image)
	end if
	
	return 0
end function

	
	



